/*
 * This file is part of IPAACA, the
 *  "Incremental Processing Architecture
 *   for Artificial Conversational Agents".
 *
 * Copyright (c) 2009-2019 Social Cognitive Systems Group
 *                         (formerly the Sociable Agents Group)
 *                         CITEC, Bielefeld University
 *
 * http://opensource.cit-ec.de/projects/ipaaca/
 * http://purl.org/net/ipaaca
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 */

/**
 * \file   ipaaca-config.cc
 *
 * \brief Source file for IPAACA configuration handling
 *
 * \author Ramin Yaghoubzadeh Torky (ryaghoubzadeh@uni-bielefeld.de)
 * \date   January, 2019
 */

#include <ipaaca/ipaaca.h>

extern char **environ;

namespace ipaaca {

template<>
int Config::get_with_default_internal<int>(const std::string& key, int const& default_value, bool warn)
{
    auto it = _data.find(key);
    if (it==_data.end()) {
        config_key_not_found(key, default_value, warn);
        return default_value;
    }
    std::size_t processed;
    int res = std::stoi(it->second, &processed);
    if (processed != it->second.size()) {
        config_conversion_failed(key, default_value);
        return default_value;
    }
    return res;
}
template<>
std::string Config::get_with_default_internal<std::string>(const std::string& key, std::string const& default_value, bool warn)
{
    auto it = _data.find(key);
    if (it==_data.end()) {
        config_key_not_found(key, default_value, warn);
        return default_value;
    }
    return it->second;
}

Config::ptr get_global_config(bool auto_parse_on_demand){
    static bool first = true;
    static Config::ptr global_config = std::make_shared<Config>();
    if (first) {
        first = false;
        if (auto_parse_on_demand) {
            IPAACA_DEBUG("Populating global configuration from default sources")
            global_config->populate_from_global_sources();
        }
    }
    return global_config;
}
void Config::populate_from_global_sources()
{
    _messages_delivered.clear(); // message variable warnings
    populate_from_environment();
}
void Config::populate_from_environment()
{
    int i = 1;
    char* envline = *environ;
    while (envline) {
        if(strncmp(envline, "IPAACA_", 7) == 0) {
            if (strlen(envline) > 1023) {
                IPAACA_ERROR("Ignoring overly long environment entry starting with IPAACA_")
            } else {
                bool good = true;
                std::string s(envline);
                std::string key("");
                std::string value("");
                size_t pos = 7;
                for (pos=7; pos<s.size(); ++pos) {
                    auto c = std::tolower(s[pos]);
                    if (c=='=') {
                        value = ipaaca::str_trim(s.substr(pos+1));
                        break;
                    } else if (c=='_') {
                        key += '.';
                    } else if ((c>='a')&&(c<='z')) {
                        key += c;
                    } else {
                        good = false;
                        break;
                    }
                }
                if (good) {
                    IPAACA_INFO("Configuration set from environment: " << key << "=\"" << value << "\"");
                    _data[key] = value;
                } else {
                    IPAACA_ERROR("Omitting a malformed environment variable " << s.substr(0, pos+1) << "...")
                }
            }
        }
        envline = *(environ+i);
        i++;
    }
}

} // of namespace ipaaca


